import logging
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from functools import lru_cache
from itertools import groupby
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    MutableMapping,
    Optional,
    Set,
    Tuple,
    Union,
)

# This import verifies that the dependencies are available.
import teradatasqlalchemy.types as custom_types
from pydantic.fields import Field
from sqlalchemy import create_engine, inspect
from sqlalchemy.engine import Engine
from sqlalchemy.engine.base import Connection
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.sql.expression import text
from teradatasqlalchemy.dialect import TeradataDialect
from teradatasqlalchemy.options import configure

from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.time_window_config import BaseTimeWindowConfig
from datahub.emitter.sql_parsing_builder import SqlParsingBuilder
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source_helpers import auto_lowercase_urns
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.source.sql.sql_common import SqlWorkUnit, register_custom_type
from datahub.ingestion.source.sql.sql_config import SQLCommonConfig
from datahub.ingestion.source.sql.sql_generic_profiler import ProfilingSqlReport
from datahub.ingestion.source.sql.two_tier_sql_source import (
    TwoTierSQLAlchemyConfig,
    TwoTierSQLAlchemySource,
)
from datahub.ingestion.source.usage.usage_common import BaseUsageConfig
from datahub.ingestion.source_report.ingestion_stage import IngestionStageReport
from datahub.ingestion.source_report.time_window import BaseTimeWindowReport
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    BytesTypeClass,
    TimeTypeClass,
)
from datahub.metadata.schema_classes import SchemaMetadataClass
from datahub.sql_parsing.schema_resolver import SchemaResolver
from datahub.sql_parsing.sqlglot_lineage import sqlglot_lineage

logger: logging.Logger = logging.getLogger(__name__)

register_custom_type(custom_types.JSON, BytesTypeClass)
register_custom_type(custom_types.INTERVAL_DAY, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_DAY_TO_SECOND, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_DAY_TO_MINUTE, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_DAY_TO_HOUR, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_SECOND, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_MINUTE, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_MINUTE_TO_SECOND, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_HOUR, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_HOUR_TO_MINUTE, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_HOUR_TO_SECOND, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_MONTH, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_YEAR, TimeTypeClass)
register_custom_type(custom_types.INTERVAL_YEAR_TO_MONTH, TimeTypeClass)
register_custom_type(custom_types.MBB, BytesTypeClass)
register_custom_type(custom_types.MBR, BytesTypeClass)
register_custom_type(custom_types.GEOMETRY, BytesTypeClass)
register_custom_type(custom_types.TDUDT, BytesTypeClass)
register_custom_type(custom_types.XML, BytesTypeClass)
register_custom_type(custom_types.PERIOD_TIME, TimeTypeClass)
register_custom_type(custom_types.PERIOD_DATE, TimeTypeClass)
register_custom_type(custom_types.PERIOD_TIMESTAMP, TimeTypeClass)


@dataclass
class TeradataTable:
    database: str
    name: str
    description: Optional[str]
    object_type: str
    create_timestamp: datetime
    last_alter_name: Optional[str]
    last_alter_timestamp: Optional[datetime]
    request_text: Optional[str]


# lru cache is set to 1 which work only in single threaded environment but it keeps the memory footprint lower
@lru_cache(maxsize=1)
def get_schema_columns(
    self: Any, connection: Connection, dbc_columns: str, schema: str
) -> Dict[str, List[Any]]:
    columns: Dict[str, List[Any]] = {}
    columns_query = f"select * from dbc.{dbc_columns} where DatabaseName (NOT CASESPECIFIC) = '{schema}' (NOT CASESPECIFIC) order by TableName, ColumnId"
    rows = connection.execute(text(columns_query)).fetchall()
    for row in rows:
        row_mapping = row._mapping
        if row_mapping.TableName not in columns:
            columns[row_mapping.TableName] = []

        columns[row_mapping.TableName].append(row_mapping)

    return columns


# lru cache is set to 1 which work only in single threaded environment but it keeps the memory footprint lower
@lru_cache(maxsize=1)
def get_schema_pk_constraints(
    self: Any, connection: Connection, schema: str
) -> Dict[str, List[Any]]:
    dbc_indices = "IndicesV" + "X" if configure.usexviews else "IndicesV"
    primary_keys: Dict[str, List[Any]] = {}
    stmt = f"select * from dbc.{dbc_indices} where DatabaseName (NOT CASESPECIFIC) = '{schema}' (NOT CASESPECIFIC) and IndexType = 'K' order by IndexNumber"
    rows = connection.execute(text(stmt)).fetchall()
    for row in rows:
        row_mapping = row._mapping
        if row_mapping.TableName not in primary_keys:
            primary_keys[row_mapping.TableName] = []

        primary_keys[row_mapping.TableName].append(row_mapping)

    return primary_keys


def optimized_get_pk_constraint(
    self: Any,
    connection: Connection,
    table_name: str,
    schema: Optional[str] = None,
    **kw: Dict[str, Any],
) -> Dict:
    """
    Override
    TODO: Check if we need PRIMARY Indices or PRIMARY KEY Indices
    TODO: Check for border cases (No PK Indices)
    """

    if schema is None:
        schema = self.default_schema_name

    # Default value for 'usexviews' is False so use dbc.IndicesV by default
    # dbc_indices = self.__get_xviews_obj("IndicesV")

    # table_obj = table(
    #    dbc_indices, column("ColumnName"), column("IndexName"), schema="dbc"
    # )

    res = []
    pk_keys = self.get_schema_pk_constraints(connection, schema)
    res = pk_keys.get(table_name, [])

    index_columns = list()
    index_name = None

    for index_column in res:
        index_columns.append(self.normalize_name(index_column.ColumnName))
        index_name = self.normalize_name(
            index_column.IndexName
        )  # There should be just one IndexName

    return {"constrained_columns": index_columns, "name": index_name}


def optimized_get_columns(
    self: Any,
    connection: Connection,
    table_name: str,
    schema: Optional[str] = None,
    tables_cache: MutableMapping[str, List[TeradataTable]] = {},
    use_qvci: bool = False,
    **kw: Dict[str, Any],
) -> List[Dict]:
    if schema is None:
        schema = self.default_schema_name

    # Using 'help schema.table.*' statements has been considered.
    # The DBC.ColumnsV provides the default value which is not available
    # with the 'help column' commands result.

    td_table: Optional[TeradataTable] = None
    # Check if the object is a view
    for t in tables_cache[schema]:
        if t.name == table_name:
            td_table = t
            break

    if td_table is None:
        logger.warning(
            f"Table {table_name} not found in cache for schema {schema}, not getting columns"
        )
        return []

    res = []
    if td_table.object_type == "View" and not use_qvci:
        # Volatile table definition is not stored in the dictionary.
        # We use the 'help schema.table.*' command instead to get information for all columns.
        # We have to do the same for views since we need the type information
        # which is not available in dbc.ColumnsV.
        res = self._get_column_help(connection, schema, table_name, column_name=None)

        # If this is a view, get types for individual columns (dbc.ColumnsV won't have types for view columns).
        # For a view or a volatile table, we have to set the default values as the 'help' command does not have it.
        col_info_list = []
        for r in res:
            updated_column_info_dict = self._update_column_help_info(r._mapping)
            col_info_list.append(dict(r._mapping, **(updated_column_info_dict)))
        res = col_info_list
    else:
        # Default value for 'usexviews' is False so use dbc.ColumnsV by default
        dbc_columns = "columnsQV" if use_qvci else "columnsV"
        dbc_columns = dbc_columns + "X" if configure.usexviews else dbc_columns
        res = self.get_schema_columns(connection, dbc_columns, schema).get(
            table_name, []
        )

    final_column_info = []
    # Don't care about ART tables now
    # Ignore the non-functional column in a PTI table
    for row in res:
        col_info = self._get_column_info(row)
        if "TSColumnType" in col_info and col_info["TSColumnType"] is not None:
            if (
                col_info["ColumnName"] == "TD_TIMEBUCKET"
                and col_info["TSColumnType"].strip() == "TB"
            ):
                continue
        final_column_info.append(col_info)

    return final_column_info


# lru cache is set to 1 which work only in single threaded environment but it keeps the memory footprint lower
@lru_cache(maxsize=1)
def get_schema_foreign_keys(
    self: Any, connection: Connection, schema: str
) -> Dict[str, List[Any]]:
    dbc_child_parent_table = (
        "All_RI_ChildrenV" + "X" if configure.usexviews else "All_RI_ChildrenV"
    )
    foreign_keys: Dict[str, List[Any]] = {}
    stmt = f"""
    SELECT dbc."All_RI_ChildrenV"."ChildDB",  dbc."All_RI_ChildrenV"."ChildTable", dbc."All_RI_ChildrenV"."IndexID", dbc."{dbc_child_parent_table}"."IndexName", dbc."{dbc_child_parent_table}"."ChildKeyColumn", dbc."{dbc_child_parent_table}"."ParentDB", dbc."{dbc_child_parent_table}"."ParentTable", dbc."{dbc_child_parent_table}"."ParentKeyColumn"
        FROM dbc."{dbc_child_parent_table}"
    WHERE ChildDB = '{schema}' ORDER BY "IndexID" ASC
    """
    rows = connection.execute(text(stmt)).fetchall()
    for row in rows:
        row_mapping = row._mapping
        if row_mapping.ChildTable not in foreign_keys:
            foreign_keys[row_mapping.ChildTable] = []

        foreign_keys[row_mapping.ChildTable].append(row_mapping)

    return foreign_keys


def optimized_get_foreign_keys(self, connection, table_name, schema=None, **kw):
    """
    Overrides base class method
    """

    if schema is None:
        schema = self.default_schema_name
    # Default value for 'usexviews' is False so use DBC.All_RI_ChildrenV by default
    res = self.get_schema_foreign_keys(connection, schema).get(table_name, [])

    def grouper(fk_row):
        return {
            "name": fk_row.IndexName or fk_row.IndexID,  # ID if IndexName is None
            "schema": fk_row.ParentDB,
            "table": fk_row.ParentTable,
        }

    # TODO: Check if there's a better way
    fk_dicts = list()
    for constraint_info, constraint_cols in groupby(res, grouper):
        fk_dict = {
            "name": str(constraint_info["name"]),
            "constrained_columns": list(),
            "referred_table": constraint_info["table"],
            "referred_schema": constraint_info["schema"],
            "referred_columns": list(),
        }

        for constraint_col in constraint_cols:
            fk_dict["constrained_columns"].append(
                self.normalize_name(constraint_col.ChildKeyColumn)
            )
            fk_dict["referred_columns"].append(
                self.normalize_name(constraint_col.ParentKeyColumn)
            )

        fk_dicts.append(fk_dict)

    return fk_dicts


def optimized_get_view_definition(
    self: Any,
    connection: Connection,
    view_name: str,
    schema: Optional[str] = None,
    tables_cache: MutableMapping[str, List[TeradataTable]] = {},
    **kw: Dict[str, Any],
) -> Optional[str]:
    if schema is None:
        schema = self.default_schema_name

    if schema not in tables_cache:
        return None

    for table in tables_cache[schema]:
        if table.name == view_name:
            return self.normalize_name(table.request_text)

    return None


@dataclass
class TeradataReport(ProfilingSqlReport, IngestionStageReport, BaseTimeWindowReport):
    num_queries_parsed: int = 0
    num_view_ddl_parsed: int = 0
    num_table_parse_failures: int = 0


class BaseTeradataConfig(TwoTierSQLAlchemyConfig):
    scheme: str = Field(default="teradatasql", description="database scheme")


class TeradataConfig(BaseTeradataConfig, BaseTimeWindowConfig):
    databases: Optional[List[str]] = Field(
        default=None,
        description=(
            "List of databases to ingest. If not specified, all databases will be ingested."
            " Even if this is specified, databases will still be filtered by `database_pattern`."
        ),
    )

    database_pattern = Field(
        default=AllowDenyPattern(
            deny=[
                "All",
                "Crashdumps",
                "Default",
                "DemoNow_Monitor",
                "EXTUSER",
                "External_AP",
                "GLOBAL_FUNCTIONS",
                "LockLogShredder",
                "PUBLIC",
                "SQLJ",
                "SYSBAR",
                "SYSJDBC",
                "SYSLIB",
                "SYSSPATIAL",
                "SYSUDTLIB",
                "SYSUIF",
                "SysAdmin",
                "Sys_Calendar",
                "SystemFe",
                "TDBCMgmt",
                "TDMaps",
                "TDPUSER",
                "TDQCD",
                "TDStats",
                "TD_ANALYTICS_DB",
                "TD_SERVER_DB",
                "TD_SYSFNLIB",
                "TD_SYSGPL",
                "TD_SYSXML",
                "TDaaS_BAR",
                "TDaaS_DB",
                "TDaaS_Maint",
                "TDaaS_Monitor",
                "TDaaS_Support",
                "TDaaS_TDBCMgmt1",
                "TDaaS_TDBCMgmt2",
                "dbcmngr",
                "mldb",
                "system",
                "tapidb",
                "tdwm",
                "val",
                "dbc",
            ]
        ),
        description="Regex patterns for databases to filter in ingestion.",
    )
    include_table_lineage = Field(
        default=False,
        description="Whether to include table lineage in the ingestion. "
        "This requires to have the table lineage feature enabled.",
    )

    include_view_lineage = Field(
        default=True,
        description="Whether to include view lineage in the ingestion. "
        "This requires to have the view lineage feature enabled.",
    )
    usage: BaseUsageConfig = Field(
        description="The usage config to use when generating usage statistics",
        default=BaseUsageConfig(),
    )

    default_db: Optional[str] = Field(
        default=None,
        description="The default database to use for unqualified table names",
    )

    include_usage_statistics: bool = Field(
        default=False,
        description="Generate usage statistic.",
    )

    use_file_backed_cache: bool = Field(
        default=True,
        description="Whether to use a file backed cache for the view definitions.",
    )

    use_qvci: bool = Field(
        default=False,
        description="Whether to use QVCI to get column information. This is faster but requires to have QVCI enabled.",
    )


@platform_name("Teradata")
@config_class(TeradataConfig)
@support_status(SupportStatus.TESTING)
@capability(SourceCapability.DOMAINS, "Enabled by default")
@capability(SourceCapability.CONTAINERS, "Enabled by default")
@capability(SourceCapability.PLATFORM_INSTANCE, "Enabled by default")
@capability(SourceCapability.DELETION_DETECTION, "Optionally enabled via configuration")
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration")
@capability(SourceCapability.LINEAGE_COARSE, "Optionally enabled via configuration")
@capability(SourceCapability.LINEAGE_FINE, "Optionally enabled via configuration")
@capability(SourceCapability.USAGE_STATS, "Optionally enabled via configuration")
class TeradataSource(TwoTierSQLAlchemySource):
    """
    This plugin extracts the following:

    - Metadata for databases, schemas, views, and tables
    - Column types associated with each table
    - Table, row, and column statistics via optional SQL profiling
    """

    config: TeradataConfig

    LINEAGE_QUERY_DATABASE_FILTER: str = """and default_database IN ({databases})"""

    LINEAGE_TIMESTAMP_BOUND_QUERY: str = """
    SELECT MIN(CollectTimeStamp) as "min_ts", MAX(CollectTimeStamp) as "max_ts" from DBC.QryLogV
    """.strip()

    QUERY_TEXT_QUERY: str = """
    SELECT
        s.QueryID as "query_id",
        UserName as "user",
        StartTime AT TIME ZONE 'GMT' as "timestamp",
        DefaultDatabase as default_database,
        s.SqlTextInfo as "query_text",
        s.SqlRowNo as "row_no"
    FROM "DBC".QryLogV as l
    JOIN "DBC".QryLogSqlV as s on s.QueryID = l.QueryID
    WHERE
        l.ErrorCode = 0
        AND l.statementtype not in (
        'Unrecognized type',
        'Create Database/User',
        'Help',
        'Modify Database',
        'Drop Table',
        'Show',
        'Not Applicable',
        'Grant',
        'Abort',
        'Database',
        'Flush Query Logging',
        'Null',
        'Begin/End DBQL',
        'Revoke'
    )
        and "timestamp" >= TIMESTAMP '{start_time}'
        and "timestamp" < TIMESTAMP '{end_time}'
        and s.CollectTimeStamp >= TIMESTAMP '{start_time}'
        and default_database not in ('DEMONOW_MONITOR')
        {databases_filter}
    ORDER BY "query_id", "row_no"
    """.strip()

    TABLES_AND_VIEWS_QUERY: str = """
SELECT
    t.DataBaseName,
    t.TableName as name,
    t.CommentString as description,
    CASE t.TableKind
         WHEN 'I' THEN 'Join index'
         WHEN 'N' THEN 'Hash index'
         WHEN 'T' THEN 'Table'
         WHEN 'V' THEN 'View'
         WHEN 'O' THEN 'NoPI Table'
         WHEN 'Q' THEN 'Queue table'
    END AS object_type,
    t.CreateTimeStamp,
    t.LastAlterName,
    t.LastAlterTimeStamp,
    t.RequestText
FROM dbc.TablesV t
WHERE DataBaseName NOT IN (
                'All',
                'Crashdumps',
                'Default',
                'DemoNow_Monitor',
                'EXTUSER',
                'External_AP',
                'GLOBAL_FUNCTIONS',
                'LockLogShredder',
                'PUBLIC',
                'SQLJ',
                'SYSBAR',
                'SYSJDBC',
                'SYSLIB',
                'SYSSPATIAL',
                'SYSUDTLIB',
                'SYSUIF',
                'SysAdmin',
                'Sys_Calendar',
                'SystemFe',
                'TDBCMgmt',
                'TDMaps',
                'TDPUSER',
                'TDQCD',
                'TDStats',
                'TD_ANALYTICS_DB',
                'TD_SERVER_DB',
                'TD_SYSFNLIB',
                'TD_SYSGPL',
                'TD_SYSXML',
                'TDaaS_BAR',
                'TDaaS_DB',
                'TDaaS_Maint',
                'TDaaS_Monitor',
                'TDaaS_Support',
                'TDaaS_TDBCMgmt1',
                'TDaaS_TDBCMgmt2',
                'dbcmngr',
                'mldb',
                'system',
                'tapidb',
                'tdwm',
                'val',
                'dbc'
)
AND t.TableKind in ('T', 'V', 'Q', 'O')
ORDER by DataBaseName, TableName;
     """.strip()

    _tables_cache: MutableMapping[str, List[TeradataTable]] = defaultdict(list)

    def __init__(self, config: TeradataConfig, ctx: PipelineContext):
        super().__init__(config, ctx, "teradata")

        self.report: TeradataReport = TeradataReport()
        self.graph: Optional[DataHubGraph] = ctx.graph

        self.builder: SqlParsingBuilder = SqlParsingBuilder(
            usage_config=(
                self.config.usage if self.config.include_usage_statistics else None
            ),
            generate_lineage=True,
            generate_usage_statistics=self.config.include_usage_statistics,
            generate_operations=self.config.usage.include_operational_stats,
        )

        self.schema_resolver = self._init_schema_resolver()

        if self.config.include_tables or self.config.include_views:
            self.cache_tables_and_views()
            logger.info(f"Found {len(self._tables_cache)} tables and views")
            setattr(self, "loop_tables", self.cached_loop_tables)  # noqa: B010
            setattr(self, "loop_views", self.cached_loop_views)  # noqa: B010
            setattr(  # noqa: B010
                self, "get_table_properties", self.cached_get_table_properties
            )

            tables_cache = self._tables_cache
            setattr(  # noqa: B010
                TeradataDialect,
                "get_columns",
                lambda self, connection, table_name, schema=None, use_qvci=self.config.use_qvci, **kw: optimized_get_columns(
                    self,
                    connection,
                    table_name,
                    schema,
                    tables_cache=tables_cache,
                    use_qvci=use_qvci,
                    **kw,
                ),
            )

            setattr(  # noqa: B010
                TeradataDialect,
                "get_pk_constraint",
                lambda self, connection, table_name, schema=None, **kw: optimized_get_pk_constraint(
                    self, connection, table_name, schema, **kw
                ),
            )

            setattr(  # noqa: B010
                TeradataDialect,
                "get_foreign_keys",
                lambda self, connection, table_name, schema=None, **kw: optimized_get_foreign_keys(
                    self, connection, table_name, schema, **kw
                ),
            )

            setattr(  # noqa: B010
                TeradataDialect,
                "get_schema_columns",
                lambda self, connection, dbc_columns, schema: get_schema_columns(
                    self, connection, dbc_columns, schema
                ),
            )

            # Disabling the below because the cached view definition is not the view definition the column in tablesv actually holds the last statement executed against the object... not necessarily the view definition
            # setattr(  # noqa: B010
            #   TeradataDialect,
            #    "get_view_definition",
            #   lambda self, connection, view_name, schema=None, **kw: optimized_get_view_definition(
            #        self, connection, view_name, schema, tables_cache=tables_cache, **kw
            #    ),
            # )

            setattr(  # noqa: B010
                TeradataDialect,
                "get_schema_pk_constraints",
                lambda self, connection, schema: get_schema_pk_constraints(
                    self, connection, schema
                ),
            )

            setattr(  # noqa: B010
                TeradataDialect,
                "get_schema_foreign_keys",
                lambda self, connection, schema: get_schema_foreign_keys(
                    self, connection, schema
                ),
            )
        else:
            logger.info(
                "Disabling stale entity removal as tables and views are disabled"
            )
            if self.config.stateful_ingestion:
                self.config.stateful_ingestion.remove_stale_metadata = False

    @classmethod
    def create(cls, config_dict, ctx):
        config = TeradataConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def _init_schema_resolver(self) -> SchemaResolver:
        if not self.config.include_tables or not self.config.include_views:
            if self.ctx.graph:
                return self.ctx.graph.initialize_schema_resolver_from_datahub(
                    platform=self.platform,
                    platform_instance=self.config.platform_instance,
                    env=self.config.env,
                )
            else:
                logger.warning(
                    "Failed to load schema info from DataHub as DataHubGraph is missing.",
                )
        return SchemaResolver(
            platform=self.platform,
            platform_instance=self.config.platform_instance,
            env=self.config.env,
        )

    def get_inspectors(self):
        # This method can be overridden in the case that you want to dynamically
        # run on multiple databases.
        url = self.config.get_sql_alchemy_url()
        logger.debug(f"sql_alchemy_url={url}")
        engine = create_engine(url, **self.config.options)
        with engine.connect() as conn:
            inspector = inspect(conn)
            if self.config.database and self.config.database != "":
                databases = [self.config.database]
            elif self.config.databases:
                databases = self.config.databases
            else:
                databases = inspector.get_schema_names()
            for db in databases:
                if self.config.database_pattern.allowed(db):
                    # url = self.config.get_sql_alchemy_url(current_db=db)
                    # with create_engine(url, **self.config.options).connect() as conn:
                    #    inspector = inspect(conn)
                    inspector._datahub_database = db
                    yield inspector

    def get_db_name(self, inspector: Inspector) -> str:
        if hasattr(inspector, "_datahub_database"):
            return inspector._datahub_database

        engine = inspector.engine

        if engine and hasattr(engine, "url") and hasattr(engine.url, "database"):
            return str(engine.url.database).strip('"')
        else:
            raise Exception("Unable to get database name from Sqlalchemy inspector")

    def cached_loop_tables(  # noqa: C901
        self,
        inspector: Inspector,
        schema: str,
        sql_config: SQLCommonConfig,
    ) -> Iterable[Union[SqlWorkUnit, MetadataWorkUnit]]:
        setattr(  # noqa: B010
            inspector,
            "get_table_names",
            lambda schema: [
                i.name
                for i in filter(
                    lambda t: t.object_type != "View", self._tables_cache[schema]
                )
            ],
        )
        yield from super().loop_tables(inspector, schema, sql_config)

    def cached_get_table_properties(
        self, inspector: Inspector, schema: str, table: str
    ) -> Tuple[Optional[str], Dict[str, str], Optional[str]]:
        description: Optional[str] = None
        properties: Dict[str, str] = {}

        # The location cannot be fetched generically, but subclasses may override
        # this method and provide a location.
        location: Optional[str] = None

        for entry in self._tables_cache[schema]:
            if entry.name == table:
                description = entry.description
                if entry.object_type == "View" and entry.request_text:
                    properties["view_definition"] = entry.request_text
                break
        return description, properties, location

    def cached_loop_views(  # noqa: C901
        self,
        inspector: Inspector,
        schema: str,
        sql_config: SQLCommonConfig,
    ) -> Iterable[Union[SqlWorkUnit, MetadataWorkUnit]]:
        setattr(  # noqa: B010
            inspector,
            "get_view_names",
            lambda schema: [
                i.name
                for i in filter(
                    lambda t: t.object_type == "View", self._tables_cache[schema]
                )
            ],
        )
        yield from super().loop_views(inspector, schema, sql_config)

    def cache_tables_and_views(self) -> None:
        engine = self.get_metadata_engine()
        for entry in engine.execute(self.TABLES_AND_VIEWS_QUERY):
            table = TeradataTable(
                database=entry.DataBaseName.strip(),
                name=entry.name.strip(),
                description=entry.description.strip() if entry.description else None,
                object_type=entry.object_type,
                create_timestamp=entry.CreateTimeStamp,
                last_alter_name=entry.LastAlterName,
                last_alter_timestamp=entry.LastAlterTimeStamp,
                request_text=(
                    entry.RequestText.strip()
                    if entry.object_type == "View" and entry.RequestText
                    else None
                ),
            )
            if table.database not in self._tables_cache:
                self._tables_cache[table.database] = []

            self._tables_cache[table.database].append(table)

    def get_audit_log_mcps(self, urns: Set[str]) -> Iterable[MetadataWorkUnit]:
        engine = self.get_metadata_engine()
        for entry in engine.execute(self._make_lineage_query()):
            self.report.num_queries_parsed += 1
            if self.report.num_queries_parsed % 1000 == 0:
                logger.info(f"Parsed {self.report.num_queries_parsed} queries")

            yield from self.gen_lineage_from_query(
                query=entry.query_text,
                default_database=entry.default_database,
                timestamp=entry.timestamp,
                user=entry.user,
                urns=urns,
            )

    def _make_lineage_query(self) -> str:
        databases_filter = (
            ""
            if not self.config.databases
            else "and default_database in ({databases})".format(
                databases=",".join([f"'{db}'" for db in self.config.databases])
            )
        )

        query = self.QUERY_TEXT_QUERY.format(
            start_time=self.config.start_time,
            end_time=self.config.end_time,
            databases_filter=databases_filter,
        )
        return query

    def gen_lineage_from_query(
        self,
        query: str,
        default_database: Optional[str] = None,
        timestamp: Optional[datetime] = None,
        user: Optional[str] = None,
        view_urn: Optional[str] = None,
        urns: Optional[Set[str]] = None,
    ) -> Iterable[MetadataWorkUnit]:
        result = sqlglot_lineage(
            # With this clever hack we can make the query parser to not fail on queries with CASESPECIFIC
            sql=query.replace("(NOT CASESPECIFIC)", ""),
            schema_resolver=self.schema_resolver,
            default_db=None,
            default_schema=(
                default_database if default_database else self.config.default_db
            ),
        )
        if result.debug_info.table_error:
            logger.debug(
                f"Error parsing table lineage ({view_urn}):\n{result.debug_info.table_error}"
            )
            self.report.num_table_parse_failures += 1
        else:
            yield from self.builder.process_sql_parsing_result(
                result,
                query=query,
                is_view_ddl=view_urn is not None,
                query_timestamp=timestamp,
                user=f"urn:li:corpuser:{user}",
                include_urns=urns,
            )

    def get_metadata_engine(self) -> Engine:
        url = self.config.get_sql_alchemy_url()
        logger.debug(f"sql_alchemy_url={url}")
        return create_engine(url, **self.config.options)

    def get_workunits_internal(self) -> Iterable[Union[MetadataWorkUnit, SqlWorkUnit]]:
        # Add all schemas to the schema resolver
        # Sql parser operates on lowercase urns so we need to lowercase the urns
        for wu in auto_lowercase_urns(super().get_workunits_internal()):
            urn = wu.get_urn()
            schema_metadata = wu.get_aspect_of_type(SchemaMetadataClass)
            if schema_metadata:
                self.schema_resolver.add_schema_metadata(urn, schema_metadata)
            yield wu

        urns = self.schema_resolver.get_urns()
        if self.config.include_table_lineage or self.config.include_usage_statistics:
            self.report.report_ingestion_stage_start("audit log extraction")
            yield from self.get_audit_log_mcps(urns=urns)

        yield from self.builder.gen_workunits()



import functools
import threading
from typing import Callable, Dict, Hashable, Tuple, TypeVar

import cachetools
import cachetools.keys
from typing_extensions import ParamSpec

_Key = Tuple[Hashable, ...]
_F = ParamSpec("_F")
_T = TypeVar("_T")


def serialized_lru_cache(
    maxsize: int,
) -> Callable[[Callable[_F, _T]], Callable[_F, _T]]:
    """Similar to `lru_cache`, but ensures multiple calls with the same parameters are serialized.

    Calls with different parameters are allowed to proceed in parallel.

    Args:
        maxsize (int): Maximum number of entries to keep in the cache.

    Returns:
        Callable[[Callable[F, T]], Callable[F, T]]: Decorator for the function to be wrapped.
    """

    UNSET = object()

    def decorator(func: Callable[_F, _T]) -> Callable[_F, _T]:
        hits = 0
        misses = 0

        cache_lock = threading.Lock()
        cache: "cachetools.LRUCache[_Key, _T]" = cachetools.LRUCache(maxsize=maxsize)

        key_locks_lock = threading.Lock()
        key_locks: Dict[_Key, threading.Lock] = {}
        key_waiters: Dict[_Key, int] = {}

        def wrapper(*args: _F.args, **kwargs: _F.kwargs) -> _T:
            # We need a type ignore here because there's no way for us to require that
            # the args and kwargs are hashable while using ParamSpec.
            key: _Key = cachetools.keys.hashkey(*args, **{k: v for k, v in kwargs.items() if "cache_exclude" not in k})  # type: ignore

            with cache_lock:
                if key in cache:
                    nonlocal hits
                    hits += 1
                    return cache[key]

            with key_locks_lock:
                if key not in key_locks:
                    key_locks[key] = threading.Lock()
                    key_waiters[key] = 0
                lock = key_locks[key]
                key_waiters[key] += 1

            try:
                with lock:
                    # Check the cache again, in case the cache was updated by another thread.
                    result = UNSET
                    with cache_lock:
                        if key in cache:
                            hits += 1
                            return cache[key]

                    nonlocal misses
                    misses += 1
                    result = func(*args, **kwargs)

                    with cache_lock:
                        cache[key] = result
                    return result

            finally:
                with key_locks_lock:
                    key_waiters[key] -= 1
                    if key_waiters[key] == 0:
                        del key_locks[key]
                        del key_waiters[key]

        def cache_info() -> functools._CacheInfo:
            return functools._CacheInfo(
                hits=hits,
                misses=misses,
                maxsize=maxsize,
                currsize=len(cache),
            )

        # Add some extra attributes to the wrapper function. This makes it mostly compatible
        # with functools.lru_cache.
        wrapper.cache = cache  # type: ignore
        wrapper.cache_info = cache_info  # type: ignore

        return functools.update_wrapper(wrapper, func)

    return decorator



import logging
from dataclasses import dataclass, field as dataclass_field
from datetime import timedelta
from typing import Any, Dict, List, Literal, Optional, Union

import pydantic
from pydantic import root_validator, validator
from pydantic.fields import Field

from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.git import GitInfo
from datahub.configuration.source_common import EnvConfigMixin
from datahub.configuration.validate_field_rename import pydantic_renamed_field
from datahub.ingestion.source.looker.looker_config import (
    LookerCommonConfig,
    LookerConnectionDefinition,
)
from datahub.ingestion.source.looker.looker_lib_wrapper import (
    LookerAPI,
    LookerAPIConfig,
    TransportOptionsConfig,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalSourceReport,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)
from datahub.utilities.lossy_collections import LossyList

logger = logging.getLogger(__name__)

NAME: str = "name"

_BASE_PROJECT_NAME = "__BASE"

_EXPLORE_FILE_EXTENSION = ".explore.lkml"

_VIEW_FILE_EXTENSION = ".view.lkml"

_MODEL_FILE_EXTENSION = ".model.lkml"

VIEW_LANGUAGE_LOOKML: str = "lookml"

VIEW_LANGUAGE_SQL: str = "sql"

DERIVED_VIEW_SUFFIX = r".sql_table_name"

DERIVED_VIEW_PATTERN: str = r"\$\{([^}]*)\}"


@dataclass
class LookMLSourceReport(StaleEntityRemovalSourceReport):
    git_clone_latency: Optional[timedelta] = None
    models_discovered: int = 0
    models_dropped: List[str] = dataclass_field(default_factory=LossyList)
    views_discovered: int = 0
    views_dropped: List[str] = dataclass_field(default_factory=LossyList)
    views_dropped_unreachable: List[str] = dataclass_field(default_factory=LossyList)
    query_parse_attempts: int = 0
    query_parse_failures: int = 0
    query_parse_failure_views: List[str] = dataclass_field(default_factory=LossyList)
    _looker_api: Optional[LookerAPI] = None

    def report_models_scanned(self) -> None:
        self.models_discovered += 1

    def report_views_scanned(self) -> None:
        self.views_discovered += 1

    def report_models_dropped(self, model: str) -> None:
        self.models_dropped.append(model)

    def report_views_dropped(self, view: str) -> None:
        self.views_dropped.append(view)

    def report_unreachable_view_dropped(self, view: str) -> None:
        self.views_dropped_unreachable.append(view)

    def compute_stats(self) -> None:
        if self._looker_api:
            self.api_stats = self._looker_api.compute_stats()
        return super().compute_stats()


class LookMLSourceConfig(
    LookerCommonConfig, StatefulIngestionConfigBase, EnvConfigMixin
):
    git_info: Optional[GitInfo] = Field(
        None,
        description="Reference to your git location. If present, supplies handy links to your lookml on the dataset "
        "entity page.",
    )
    _github_info_deprecated = pydantic_renamed_field("github_info", "git_info")
    base_folder: Optional[pydantic.DirectoryPath] = Field(
        None,
        description="Required if not providing github configuration and deploy keys. A pointer to a local directory ("
        "accessible to the ingestion system) where the root of the LookML repo has been checked out ("
        "typically via a git clone). This is typically the root folder where the `*.model.lkml` and "
        "`*.view.lkml` files are stored. e.g. If you have checked out your LookML repo under "
        "`/Users/jdoe/workspace/my-lookml-repo`, then set `base_folder` to "
        "`/Users/jdoe/workspace/my-lookml-repo`.",
    )
    project_dependencies: Dict[str, Union[pydantic.DirectoryPath, GitInfo]] = Field(
        {},
        description="A map of project_name to local directory (accessible to the ingestion system) or Git credentials. "
        "Every local_dependencies or private remote_dependency listed in the main project's manifest.lkml file should "
        "have a corresponding entry here."
        "If a deploy key is not provided, the ingestion system will use the same deploy key as the main project. ",
    )
    connection_to_platform_map: Optional[Dict[str, LookerConnectionDefinition]] = Field(
        None,
        description="A mapping of [Looker connection names]("
        "https://docs.looker.com/reference/model-params/connection-for-model) to DataHub platform, "
        "database, and schema values.",
    )
    model_pattern: AllowDenyPattern = Field(
        AllowDenyPattern.allow_all(),
        description="List of regex patterns for LookML models to include in the extraction.",
    )
    view_pattern: AllowDenyPattern = Field(
        AllowDenyPattern.allow_all(),
        description="List of regex patterns for LookML views to include in the extraction.",
    )
    parse_table_names_from_sql: bool = Field(True, description="See note below.")
    api: Optional[LookerAPIConfig]
    project_name: Optional[str] = Field(
        None,
        description="Required if you don't specify the `api` section. The project name within which all the model "
        "files live. See (https://docs.looker.com/data-modeling/getting-started/how-project-works) to "
        "understand what the Looker project name should be. The simplest way to see your projects is to "
        "click on `Develop` followed by `Manage LookML Projects` in the Looker application.",
    )
    transport_options: Optional[TransportOptionsConfig] = Field(
        None,
        description="Populates the [TransportOptions](https://github.com/looker-open-source/sdk-codegen/blob"
        "/94d6047a0d52912ac082eb91616c1e7c379ab262/python/looker_sdk/rtl/transport.py#L70) struct for "
        "looker client",
    )
    max_file_snippet_length: int = Field(
        512000,  # 512KB should be plenty
        description="When extracting the view definition from a lookml file, the maximum number of characters to "
        "extract.",
    )
    emit_reachable_views_only: bool = Field(
        True,
        description="When enabled, only views that are reachable from explores defined in the model files are emitted",
    )
    populate_sql_logic_for_missing_descriptions: bool = Field(
        False,
        description="When enabled, field descriptions will include the sql logic for computed fields if descriptions "
        "are missing",
    )
    process_isolation_for_sql_parsing: bool = Field(
        False,
        description="When enabled, sql parsing will be executed in a separate process to prevent memory leaks.",
    )
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = Field(
        default=None, description=""
    )
    process_refinements: bool = Field(
        False,
        description="When enabled, looker refinement will be processed to adapt an existing view.",
    )

    liquid_variable: Dict[Any, Any] = Field(
        {},
        description="A dictionary containing Liquid variables and their corresponding values, utilized in SQL-defined "
        "derived views. The Liquid template will be resolved in view.derived_table.sql and "
        "view.sql_table_name. Defaults to an empty dictionary.",
    )

    looker_environment: Literal["prod", "dev"] = Field(
        "prod",
        description="A looker prod or dev environment. "
        "It helps to evaluate looker if comments i.e. -- if prod --. "
        "All if comments are evaluated to true for configured looker_environment value",
    )

    @validator("connection_to_platform_map", pre=True)
    def convert_string_to_connection_def(cls, conn_map):
        # Previous version of config supported strings in connection map. This upconverts strings to ConnectionMap
        for key in conn_map:
            if isinstance(conn_map[key], str):
                platform = conn_map[key]
                if "." in platform:
                    platform_db_split = conn_map[key].split(".")
                    connection = LookerConnectionDefinition(
                        platform=platform_db_split[0],
                        default_db=platform_db_split[1],
                        default_schema="",
                    )
                    conn_map[key] = connection
                else:
                    logger.warning(
                        f"Connection map for {key} provides platform {platform} but does not provide a default "
                        f"database name. This might result in failed resolution"
                    )
                    conn_map[key] = LookerConnectionDefinition(
                        platform=platform, default_db="", default_schema=""
                    )
        return conn_map

    @root_validator(skip_on_failure=True)
    def check_either_connection_map_or_connection_provided(cls, values):
        """Validate that we must either have a connection map or an api credential"""
        if not values.get("connection_to_platform_map", {}) and not values.get(
            "api", {}
        ):
            raise ValueError(
                "Neither api not connection_to_platform_map config was found. LookML source requires either api "
                "credentials for Looker or a map of connection names to platform identifiers to work correctly"
            )
        return values

    @root_validator(skip_on_failure=True)
    def check_either_project_name_or_api_provided(cls, values):
        """Validate that we must either have a project name or an api credential to fetch project names"""
        if not values.get("project_name") and not values.get("api"):
            raise ValueError(
                "Neither project_name not an API credential was found. LookML source requires either api credentials "
                "for Looker or a project_name to accurately name views and models."
            )
        return values

    @validator("base_folder", always=True)
    def check_base_folder_if_not_provided(
        cls, v: Optional[pydantic.DirectoryPath], values: Dict[str, Any]
    ) -> Optional[pydantic.DirectoryPath]:
        if v is None:
            git_info: Optional[GitInfo] = values.get("git_info")
            if git_info:
                if not git_info.deploy_key:
                    logger.warning(
                        "git_info is provided, but no SSH key is present. If the repo is not public, we'll fail to "
                        "clone it."
                    )
            else:
                raise ValueError("Neither base_folder nor git_info has been provided.")
        return v



from typing import List, Optional

from datahub.api.entities.assertion.field_assertion import FieldMetricAssertion
from datahub.api.entities.assertion.field_metric import FieldMetric
from datahub.integrations.assertion.common import get_entity_name


class SnowflakeFieldMetricSQLGenerator:
    def unique_count_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select count(distinct {field_name})
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def unique_percentage_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select count(distinct {field_name})/count(*)
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def null_count_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        where_clause = self._setup_where_clause(
            [dataset_filter, f"{field_name} is null"]
        )
        return f"""select count(*)
        from {entity_name} {where_clause}"""

    def null_percentage_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select ({self.null_count_sql(field_name, entity_name, dataset_filter)})/count(*)
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def min_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select min({field_name})
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def max_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select max({field_name})
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def mean_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select avg({field_name})
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def median_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select median({field_name})
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def stddev_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select stddev({field_name})
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def negative_count_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        where_clause = self._setup_where_clause([dataset_filter, f"{field_name} < 0"])
        return f"""select count(*)
        from {entity_name} {where_clause}"""

    def negative_percentage_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select ({self.negative_count_sql(field_name, entity_name, dataset_filter)})/count(*)
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def zero_count_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        where_clause = self._setup_where_clause([dataset_filter, f"{field_name} = 0"])
        return f"""select count(*)
        from {entity_name} {where_clause}"""

    def zero_percentage_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select ({self.zero_count_sql(field_name, entity_name, dataset_filter)})/count(*)
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def min_length_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select min(length({field_name}))
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def max_length_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select max(length({field_name}))
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def empty_count_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        where_clause = self._setup_where_clause(
            [dataset_filter, f"({field_name} is null or trim({field_name})='')"]
        )
        return f"""select count(*)
        from {entity_name} {where_clause}"""

    def empty_percentage_sql(
        self, field_name: str, entity_name: str, dataset_filter: Optional[str]
    ) -> str:
        return f"""select ({self.empty_count_sql(field_name, entity_name, dataset_filter)})/count(*)
        from {entity_name} {self._setup_where_clause([dataset_filter])}"""

    def _setup_where_clause(self, filters: List[Optional[str]]) -> str:
        where_clause = " and ".join(f for f in filters if f)
        return f"where {where_clause}" if where_clause else ""

    def metric_sql(self, assertion: FieldMetricAssertion) -> str:
        metric_sql_mapping = {
            FieldMetric.UNIQUE_COUNT: self.unique_count_sql,
            FieldMetric.UNIQUE_PERCENTAGE: self.unique_percentage_sql,
            FieldMetric.NULL_COUNT: self.null_count_sql,
            FieldMetric.NULL_PERCENTAGE: self.null_percentage_sql,
            FieldMetric.MIN: self.min_sql,
            FieldMetric.MAX: self.max_sql,
            FieldMetric.MEAN: self.mean_sql,
            FieldMetric.MEDIAN: self.median_sql,
            FieldMetric.STDDEV: self.stddev_sql,
            FieldMetric.NEGATIVE_COUNT: self.negative_count_sql,
            FieldMetric.NEGATIVE_PERCENTAGE: self.negative_percentage_sql,
            FieldMetric.ZERO_COUNT: self.zero_count_sql,
            FieldMetric.ZERO_PERCENTAGE: self.zero_percentage_sql,
            FieldMetric.MIN_LENGTH: self.min_length_sql,
            FieldMetric.MAX_LENGTH: self.max_length_sql,
            FieldMetric.EMPTY_COUNT: self.empty_count_sql,
            FieldMetric.EMPTY_PERCENTAGE: self.empty_percentage_sql,
        }

        entity_name = ".".join(get_entity_name(assertion))

        return metric_sql_mapping[assertion.metric](
            assertion.field,
            entity_name,
            (
                assertion.filters.sql
                if assertion.filters and assertion.filters.sql
                else None
            ),
        )



from typing import Callable, Optional, Union

from typing_extensions import Protocol

from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.metadata.com.linkedin.pegasus2avro.mxe import (
    MetadataChangeEvent,
    MetadataChangeProposal,
)


class Emitter(Protocol):
    def emit(
        self,
        item: Union[
            MetadataChangeEvent,
            MetadataChangeProposal,
            MetadataChangeProposalWrapper,
        ],
        # NOTE: This signature should have the exception be optional rather than
        #      required. However, this would be a breaking change that may need
        #      more careful consideration.
        callback: Optional[Callable[[Exception, str], None]] = None,
    ) -> None:
        raise NotImplementedError

    def flush(self) -> None:
        pass



import datetime
import os
from typing import List, Literal, Optional

import certifi
import pydantic
from pydantic import Field, validator

from datahub.configuration.common import AllowDenyPattern, ConfigModel
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.ingestion.source.ge_profiling_config import GEProfilingConfig
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)
from datahub.ingestion.source.usage.usage_common import BaseUsageConfig
from datahub.ingestion.source_config.operation_config import is_profiling_enabled


class DremioConnectionConfig(ConfigModel):
    hostname: Optional[str] = Field(
        default=None,
        description="Hostname or IP Address of the Dremio server",
    )

    port: int = Field(
        default=9047,
        description="Port of the Dremio REST API",
    )

    username: Optional[str] = Field(
        default=None,
        description="Dremio username",
    )

    authentication_method: Optional[str] = Field(
        default="PAT",
        description="Authentication method: 'password' or 'PAT' (Personal Access Token)",
    )

    password: Optional[str] = Field(
        default=None,
        description="Dremio password or Personal Access Token",
    )

    tls: bool = Field(
        default=True,
        description="Whether the Dremio REST API port is encrypted",
    )

    disable_certificate_verification: Optional[bool] = Field(
        default=False,
        description="Disable TLS certificate verification",
    )

    path_to_certificates: str = Field(
        default=certifi.where(),
        description="Path to SSL certificates",
    )

    is_dremio_cloud: bool = Field(
        default=False,
        description="Whether this is a Dremio Cloud instance",
    )

    # https://docs.dremio.com/cloud/reference/api/#api-endpoint
    dremio_cloud_region: Literal["US", "EU"] = Field(
        default="US",
        description="Dremio Cloud region ('US' or 'EU')",
    )

    dremio_cloud_project_id: Optional[str] = Field(
        default=None,
        description="ID of Dremio Cloud Project. Found in Project Settings in the Dremio Cloud UI",
    )

    @validator("authentication_method")
    def validate_auth_method(cls, value):
        allowed_methods = ["password", "PAT"]
        if value not in allowed_methods:
            raise ValueError(
                f"authentication_method must be one of {allowed_methods}",
            )
        return value

    @validator("password")
    def validate_password(cls, value, values):
        if values.get("authentication_method") == "PAT" and not value:
            raise ValueError(
                "Password (Personal Access Token) is required when using PAT authentication",
            )
        return value


class ProfileConfig(GEProfilingConfig):

    query_timeout: int = Field(
        default=300, description="Time before cancelling Dremio profiling query"
    )

    row_count: bool = True
    column_count: bool = True
    sample_values: bool = True

    # Below Configs inherited from GEProfilingConfig
    # but not used in Dremio so we hide them from docs.
    include_field_median_value: bool = Field(
        default=False,
        hidden_from_docs=True,
        description="Median causes a number of issues in Dremio.",
    )
    partition_profiling_enabled: bool = Field(default=True, hidden_from_docs=True)
    profile_table_row_count_estimate_only: bool = Field(
        default=False, hidden_from_docs=True
    )
    query_combiner_enabled: bool = Field(default=True, hidden_from_docs=True)
    max_number_of_fields_to_profile: Optional[pydantic.PositiveInt] = Field(
        default=None, hidden_from_docs=True
    )
    profile_if_updated_since_days: Optional[pydantic.PositiveFloat] = Field(
        default=None, hidden_from_docs=True
    )
    profile_table_size_limit: Optional[int] = Field(
        default=5,
        description="Profile tables only if their size is less then specified GBs. If set to `null`, no limit on the size of tables to profile. Supported only in `snowflake` and `BigQuery`",
        hidden_from_docs=True,
    )

    profile_table_row_limit: Optional[int] = Field(
        default=5000000,
        hidden_from_docs=True,
        description="Profile tables only if their row count is less then specified count. If set to `null`, no limit on the row count of tables to profile. Supported only in `snowflake` and `BigQuery`",
    )

    partition_datetime: Optional[datetime.datetime] = Field(
        default=None,
        hidden_from_docs=True,
        description="If specified, profile only the partition which matches this datetime. "
        "If not specified, profile the latest partition. Only Bigquery supports this.",
    )
    use_sampling: bool = Field(
        default=True,
        hidden_from_docs=True,
        description="Whether to profile column level stats on sample of table. Only BigQuery and Snowflake support this. "
        "If enabled, profiling is done on rows sampled from table. Sampling is not done for smaller tables. ",
    )

    sample_size: int = Field(
        default=10000,
        hidden_from_docs=True,
        description="Number of rows to be sampled from table for column level profiling."
        "Applicable only if `use_sampling` is set to True.",
    )
    profile_external_tables: bool = Field(
        default=False,
        hidden_from_docs=True,
        description="Whether to profile external tables. Only Snowflake and Redshift supports this.",
    )

    tags_to_ignore_sampling: Optional[List[str]] = pydantic.Field(
        default=None,
        hidden_from_docs=True,
        description=(
            "Fixed list of tags to ignore sampling."
            " If not specified, tables will be sampled based on `use_sampling`."
        ),
    )


class DremioSourceMapping(EnvConfigMixin, PlatformInstanceConfigMixin, ConfigModel):
    platform: str = Field(
        description="Source connection made by Dremio (e.g. S3, Snowflake)",
    )
    source_name: str = Field(
        description="Alias of platform in Dremio connection",
    )


class DremioSourceConfig(
    DremioConnectionConfig,
    StatefulIngestionConfigBase,
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
):

    domain: Optional[str] = Field(
        default=None,
        description="Domain for all source objects.",
    )

    source_mappings: Optional[List[DremioSourceMapping]] = Field(
        default=None,
        description="Mappings from Dremio sources to DataHub platforms and datasets.",
    )

    # Entity Filters
    schema_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for schemas to filter",
    )

    dataset_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for tables and views to filter in ingestion. Specify regex to match the entire table name in dremio.schema.table format. e.g. to match all tables starting with customer in Customer database and public schema, use the regex 'dremio.public.customer.*'",
    )

    usage: BaseUsageConfig = Field(
        description="The usage config to use when generating usage statistics",
        default=BaseUsageConfig(),
    )

    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = None

    # Profiling
    profile_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for tables to profile",
    )
    profiling: ProfileConfig = Field(
        default=ProfileConfig(),
        description="Configuration for profiling",
    )

    def is_profiling_enabled(self) -> bool:
        return self.profiling.enabled and is_profiling_enabled(
            self.profiling.operation_config
        )

    # Advanced Configs
    max_workers: int = Field(
        default=5 * (os.cpu_count() or 4),
        description="Number of worker threads to use for parallel processing",
    )

    include_query_lineage: bool = Field(
        default=False,
        description="Whether to include query-based lineage information.",
    )



import json
import logging

import click
from click_default_group import DefaultGroup

from datahub.ingestion.graph.client import get_default_graph
from datahub.telemetry import telemetry
from datahub.upgrade import upgrade

logger = logging.getLogger(__name__)


@click.group(cls=DefaultGroup, default="urn")
def state() -> None:
    """Managed state stored in DataHub by stateful ingestion."""
    pass


@state.command()
@click.option("--pipeline-name", required=True, type=str)
@click.option("--platform", required=True, type=str)
@upgrade.check_upgrade
@telemetry.with_telemetry()
def inspect(pipeline_name: str, platform: str) -> None:
    """
    Get the latest stateful ingestion state for a given pipeline.
    Only works for state entity removal for now.
    """

    datahub_graph = get_default_graph()
    checkpoint = datahub_graph.get_latest_pipeline_checkpoint(pipeline_name, platform)
    if not checkpoint:
        click.secho("No ingestion state found.", fg="red")
        exit(1)

    logger.info(f"Found ingestion state with {len(checkpoint.state.urns)} URNs.")
    click.echo(json.dumps(checkpoint.state.urns, indent=2))



import logging
import time
from contextlib import AbstractContextManager
from typing import Any, Optional

logger: logging.Logger = logging.getLogger(__name__)


class PerfTimer(AbstractContextManager):
    """
    A context manager that gives easy access to elapsed time for performance measurement.
    """

    def __init__(self) -> None:
        self.start_time: Optional[float] = None
        self.end_time: Optional[float] = None
        self._past_active_time: float = 0
        self.paused: bool = False
        self._error_state = False

    def start(self) -> None:
        if self.end_time is not None:
            self._past_active_time = self.elapsed_seconds()

        self.start_time = time.perf_counter()
        self.end_time = None
        self.paused = False

    def pause(self) -> "PerfTimer":
        self.assert_timer_is_running()
        self._past_active_time = self.elapsed_seconds()
        self.start_time = None
        self.end_time = None
        self.paused = True
        return self

    def finish(self) -> None:
        self.assert_timer_is_running()
        self.end_time = time.perf_counter()

    def __enter__(self) -> "PerfTimer":
        if self.paused:  # Entering paused timer context, NO OP
            pass
        else:
            self.start()
        return self

    def __exit__(
        self,
        exc_type: Any,
        exc: Any,
        traceback: Any,
    ) -> Optional[bool]:
        if self.paused:  # Exiting paused timer context, resume timer
            self.start()
        else:
            self.finish()
        return None

    def elapsed_seconds(self) -> float:
        """
        Returns the elapsed time in seconds.
        """
        if self.paused or not self.start_time:
            return self._past_active_time

        if self.end_time is None:
            return (time.perf_counter() - self.start_time) + (self._past_active_time)
        else:
            return (self.end_time - self.start_time) + self._past_active_time

    def assert_timer_is_running(self) -> None:
        """
        Returns true if timer is in running state.
        Timer is in NOT in running state if
        1. it has never been started.
        2. it is in paused state.
        3. it had been started and finished in the past but not started again.
        """
        if self.start_time is None or self.paused or self.end_time:
            self._error_state = True
            logger.warning("Did you forget to start the timer ?")

    def __repr__(self) -> str:
        return repr(self.as_obj())

    def __str__(self) -> str:
        return self.__repr__()

    def as_obj(self) -> Optional[str]:
        if self.start_time is None:
            return None
        else:
            time_taken = self.elapsed_seconds()
            state = " (error)" if self._error_state else ""
            return f"{time_taken:.3f} seconds{state}"



import json
import logging
import uuid
from typing import Any, Dict, Iterable, List, Optional

from pyiceberg.catalog import Catalog
from pyiceberg.exceptions import NoSuchIcebergTableError
from pyiceberg.schema import Schema, SchemaVisitorPerPrimitiveType, visit
from pyiceberg.table import Table
from pyiceberg.typedef import Identifier
from pyiceberg.types import (
    BinaryType,
    BooleanType,
    DateType,
    DecimalType,
    DoubleType,
    FixedType,
    FloatType,
    IntegerType,
    ListType,
    LongType,
    MapType,
    NestedField,
    StringType,
    StructType,
    TimestampType,
    TimestamptzType,
    TimeType,
    UUIDType,
)

from datahub.emitter.mce_builder import (
    make_data_platform_urn,
    make_dataplatform_instance_urn,
    make_dataset_urn_with_platform_instance,
    make_group_urn,
    make_user_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.extractor import schema_util
from datahub.ingestion.source.iceberg.iceberg_common import (
    IcebergSourceConfig,
    IcebergSourceReport,
)
from datahub.ingestion.source.iceberg.iceberg_profiler import IcebergProfiler
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import Status
from datahub.metadata.com.linkedin.pegasus2avro.metadata.snapshot import DatasetSnapshot
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    OtherSchema,
    SchemaField,
    SchemaMetadata,
)
from datahub.metadata.schema_classes import (
    DataPlatformInstanceClass,
    DatasetPropertiesClass,
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
)

LOGGER = logging.getLogger(__name__)
logging.getLogger("azure.core.pipeline.policies.http_logging_policy").setLevel(
    logging.WARNING
)


@platform_name("Iceberg")
@support_status(SupportStatus.TESTING)
@config_class(IcebergSourceConfig)
@capability(
    SourceCapability.PLATFORM_INSTANCE,
    "Optionally enabled via configuration, an Iceberg instance represents the catalog name where the table is stored.",
)
@capability(SourceCapability.DOMAINS, "Currently not supported.", supported=False)
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration.")
@capability(
    SourceCapability.PARTITION_SUPPORT, "Currently not supported.", supported=False
)
@capability(SourceCapability.DESCRIPTIONS, "Enabled by default.")
@capability(
    SourceCapability.OWNERSHIP,
    "Optionally enabled via configuration by specifying which Iceberg table property holds user or group ownership.",
)
@capability(SourceCapability.DELETION_DETECTION, "Enabled via stateful ingestion")
class IcebergSource(StatefulIngestionSourceBase):
    """
    ## Integration Details

    The DataHub Iceberg source plugin extracts metadata from [Iceberg tables](https://iceberg.apache.org/spec/) stored in a distributed or local file system.
    Typically, Iceberg tables are stored in a distributed file system like S3 or Azure Data Lake Storage (ADLS) and registered in a catalog.  There are various catalog
    implementations like Filesystem-based, RDBMS-based or even REST-based catalogs.  This Iceberg source plugin relies on the
    [pyiceberg library](https://py.iceberg.apache.org/).
    """

    def __init__(self, config: IcebergSourceConfig, ctx: PipelineContext) -> None:
        super().__init__(config, ctx)
        self.platform: str = "iceberg"
        self.report: IcebergSourceReport = IcebergSourceReport()
        self.config: IcebergSourceConfig = config

    @classmethod
    def create(cls, config_dict: Dict, ctx: PipelineContext) -> "IcebergSource":
        config = IcebergSourceConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def _get_datasets(self, catalog: Catalog) -> Iterable[Identifier]:
        for namespace in catalog.list_namespaces():
            yield from catalog.list_tables(namespace)

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        try:
            catalog = self.config.get_catalog()
        except Exception as e:
            LOGGER.error("Failed to get catalog", exc_info=True)
            self.report.report_failure("get-catalog", f"Failed to get catalog: {e}")
            return

        for dataset_path in self._get_datasets(catalog):
            dataset_name = ".".join(dataset_path)
            if not self.config.table_pattern.allowed(dataset_name):
                # Dataset name is rejected by pattern, report as dropped.
                self.report.report_dropped(dataset_name)
                continue

            try:
                # Try to load an Iceberg table.  Might not contain one, this will be caught by NoSuchIcebergTableError.
                table = catalog.load_table(dataset_path)
                yield from self._create_iceberg_workunit(dataset_name, table)
            except NoSuchIcebergTableError as e:
                self.report.report_failure("general", f"Failed to create workunit: {e}")
                LOGGER.exception(
                    f"Exception while processing table {dataset_path}, skipping it.",
                )

    def _create_iceberg_workunit(
        self, dataset_name: str, table: Table
    ) -> Iterable[MetadataWorkUnit]:
        self.report.report_table_scanned(dataset_name)
        dataset_urn: str = make_dataset_urn_with_platform_instance(
            self.platform,
            dataset_name,
            self.config.platform_instance,
            self.config.env,
        )
        dataset_snapshot = DatasetSnapshot(
            urn=dataset_urn,
            aspects=[Status(removed=False)],
        )

        # Dataset properties aspect.
        custom_properties = table.metadata.properties.copy()
        custom_properties["location"] = table.metadata.location
        custom_properties["format-version"] = str(table.metadata.format_version)
        custom_properties["partition-spec"] = str(self._get_partition_aspect(table))
        if table.current_snapshot():
            custom_properties["snapshot-id"] = str(table.current_snapshot().snapshot_id)
            custom_properties["manifest-list"] = table.current_snapshot().manifest_list
        dataset_properties = DatasetPropertiesClass(
            name=table.name()[-1],
            tags=[],
            description=table.metadata.properties.get("comment", None),
            customProperties=custom_properties,
        )
        dataset_snapshot.aspects.append(dataset_properties)

        # Dataset ownership aspect.
        dataset_ownership = self._get_ownership_aspect(table)
        if dataset_ownership:
            dataset_snapshot.aspects.append(dataset_ownership)

        schema_metadata = self._create_schema_metadata(dataset_name, table)
        dataset_snapshot.aspects.append(schema_metadata)

        mce = MetadataChangeEvent(proposedSnapshot=dataset_snapshot)
        yield MetadataWorkUnit(id=dataset_name, mce=mce)

        dpi_aspect = self._get_dataplatform_instance_aspect(dataset_urn=dataset_urn)
        if dpi_aspect:
            yield dpi_aspect

        if self.config.is_profiling_enabled():
            profiler = IcebergProfiler(self.report, self.config.profiling)
            yield from profiler.profile_table(dataset_name, dataset_urn, table)

    def _get_partition_aspect(self, table: Table) -> Optional[str]:
        """Extracts partition information from the provided table and returns a JSON array representing the [partition spec](https://iceberg.apache.org/spec/?#partition-specs) of the table.
        Each element of the returned array represents a field in the [partition spec](https://iceberg.apache.org/spec/?#partition-specs) that follows [Appendix-C](https://iceberg.apache.org/spec/?#appendix-c-json-serialization) of the Iceberg specification.
        Extra information has been added to this spec to make the information more user-friendly.

        Since Datahub does not have a place in its model to store this information, it is saved as a JSON string and displayed as a table property.

        Here is an example:
        ```json
        "partition-spec": "[{\"name\": \"timeperiod_loaded\", \"transform\": \"identity\", \"source\": \"timeperiod_loaded\", \"source-id\": 19, \"source-type\": \"date\", \"field-id\": 1000}]",
        ```

        Args:
            table (Table): The Iceberg table to extract partition spec from.

        Returns:
            str: JSON representation of the partition spec of the provided table (empty array if table is not partitioned) or `None` if an error occured.
        """
        try:
            return json.dumps(
                [
                    {
                        "name": partition.name,
                        "transform": str(partition.transform),
                        "source": str(
                            table.schema().find_column_name(partition.source_id)
                        ),
                        "source-id": partition.source_id,
                        "source-type": str(
                            table.schema().find_type(partition.source_id)
                        ),
                        "field-id": partition.field_id,
                    }
                    for partition in table.spec().fields
                ]
            )
        except Exception as e:
            self.report.report_warning(
                "extract-partition",
                f"Failed to extract partition spec from Iceberg table {table.name()} due to error: {str(e)}",
            )
            return None

    def _get_ownership_aspect(self, table: Table) -> Optional[OwnershipClass]:
        owners = []
        if self.config.user_ownership_property:
            if self.config.user_ownership_property in table.metadata.properties:
                user_owner = table.metadata.properties[
                    self.config.user_ownership_property
                ]
                owners.append(
                    OwnerClass(
                        owner=make_user_urn(user_owner),
                        type=OwnershipTypeClass.TECHNICAL_OWNER,
                        source=None,
                    )
                )
        if self.config.group_ownership_property:
            if self.config.group_ownership_property in table.metadata.properties:
                group_owner = table.metadata.properties[
                    self.config.group_ownership_property
                ]
                owners.append(
                    OwnerClass(
                        owner=make_group_urn(group_owner),
                        type=OwnershipTypeClass.TECHNICAL_OWNER,
                        source=None,
                    )
                )
        return OwnershipClass(owners=owners) if owners else None

    def _get_dataplatform_instance_aspect(
        self, dataset_urn: str
    ) -> Optional[MetadataWorkUnit]:
        # If we are a platform instance based source, emit the instance aspect
        if self.config.platform_instance:
            return MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=DataPlatformInstanceClass(
                    platform=make_data_platform_urn(self.platform),
                    instance=make_dataplatform_instance_urn(
                        self.platform, self.config.platform_instance
                    ),
                ),
            ).as_workunit()

        return None

    def _create_schema_metadata(
        self, dataset_name: str, table: Table
    ) -> SchemaMetadata:
        schema_fields = self._get_schema_fields_for_schema(table.schema())
        schema_metadata = SchemaMetadata(
            schemaName=dataset_name,
            platform=make_data_platform_urn(self.platform),
            version=0,
            hash="",
            platformSchema=OtherSchema(rawSchema=str(table.schema())),
            fields=schema_fields,
        )
        return schema_metadata

    def _get_schema_fields_for_schema(
        self,
        schema: Schema,
    ) -> List[SchemaField]:
        avro_schema = visit(schema, ToAvroSchemaIcebergVisitor())
        schema_fields = schema_util.avro_schema_to_mce_fields(
            json.dumps(avro_schema), default_nullable=False
        )
        return schema_fields

    def get_report(self) -> SourceReport:
        return self.report


class ToAvroSchemaIcebergVisitor(SchemaVisitorPerPrimitiveType[Dict[str, Any]]):
    """Implementation of a visitor to build an Avro schema as a dictionary from an Iceberg schema."""

    @staticmethod
    def _gen_name(prefix: str) -> str:
        return f"{prefix}{str(uuid.uuid4()).replace('-', '')}"

    def schema(self, schema: Schema, struct_result: Dict[str, Any]) -> Dict[str, Any]:
        return struct_result

    def struct(
        self, struct: StructType, field_results: List[Dict[str, Any]]
    ) -> Dict[str, Any]:
        nullable = True
        return {
            "type": "record",
            "name": self._gen_name("__struct_"),
            "fields": field_results,
            "native_data_type": str(struct),
            "_nullable": nullable,
        }

    def field(self, field: NestedField, field_result: Dict[str, Any]) -> Dict[str, Any]:
        field_result["_nullable"] = not field.required
        return {
            "name": field.name,
            "type": field_result,
            "doc": field.doc,
        }

    def list(
        self, list_type: ListType, element_result: Dict[str, Any]
    ) -> Dict[str, Any]:
        return {
            "type": "array",
            "items": element_result,
            "native_data_type": str(list_type),
            "_nullable": not list_type.element_required,
        }

    def map(
        self,
        map_type: MapType,
        key_result: Dict[str, Any],
        value_result: Dict[str, Any],
    ) -> Dict[str, Any]:
        # The Iceberg Map type will be handled differently.  The idea is to translate the map
        # similar to the Map.Entry struct of Java i.e. as an array of map_entry struct, where
        # the map_entry struct has a key field and a value field. The key and value type can
        # be complex or primitive types.
        key_result["_nullable"] = False
        value_result["_nullable"] = not map_type.value_required
        map_entry = {
            "type": "record",
            "name": self._gen_name("__map_entry_"),
            "fields": [
                {
                    "name": "key",
                    "type": key_result,
                },
                {
                    "name": "value",
                    "type": value_result,
                },
            ],
        }
        return {
            "type": "array",
            "items": map_entry,
            "native_data_type": str(map_type),
        }

    def visit_fixed(self, fixed_type: FixedType) -> Dict[str, Any]:
        return {
            "type": "fixed",
            "name": self._gen_name("__fixed_"),
            "size": len(fixed_type),
            "native_data_type": str(fixed_type),
        }

    def visit_decimal(self, decimal_type: DecimalType) -> Dict[str, Any]:
        # Also of interest: https://avro.apache.org/docs/current/spec.html#Decimal
        return {
            # "type": "bytes", # when using bytes, avro drops _nullable attribute and others.  See unit test.
            "type": "fixed",  # to fix avro bug ^ resolved by using a fixed type
            "name": self._gen_name(
                "__fixed_"
            ),  # to fix avro bug ^ resolved by using a fixed type
            "size": 1,  # to fix avro bug ^ resolved by using a fixed type
            "logicalType": "decimal",
            "precision": decimal_type.precision,
            "scale": decimal_type.scale,
            "native_data_type": str(decimal_type),
        }

    def visit_boolean(self, boolean_type: BooleanType) -> Dict[str, Any]:
        return {
            "type": "boolean",
            "native_data_type": str(boolean_type),
        }

    def visit_integer(self, integer_type: IntegerType) -> Dict[str, Any]:
        return {
            "type": "int",
            "native_data_type": str(integer_type),
        }

    def visit_long(self, long_type: LongType) -> Dict[str, Any]:
        return {
            "type": "long",
            "native_data_type": str(long_type),
        }

    def visit_float(self, float_type: FloatType) -> Dict[str, Any]:
        return {
            "type": "float",
            "native_data_type": str(float_type),
        }

    def visit_double(self, double_type: DoubleType) -> Dict[str, Any]:
        return {
            "type": "double",
            "native_data_type": str(double_type),
        }

    def visit_date(self, date_type: DateType) -> Dict[str, Any]:
        return {
            "type": "int",
            "logicalType": "date",
            "native_data_type": str(date_type),
        }

    def visit_time(self, time_type: TimeType) -> Dict[str, Any]:
        return {
            "type": "long",
            "logicalType": "time-micros",
            "native_data_type": str(time_type),
        }

    def visit_timestamp(self, timestamp_type: TimestampType) -> Dict[str, Any]:
        # Avro supports 2 types of timestamp:
        #  - Timestamp: independent of a particular timezone or calendar (TZ information is lost)
        #  - Local Timestamp: represents a timestamp in a local timezone, regardless of what specific time zone is considered local
        # utcAdjustment: bool = True
        return {
            "type": "long",
            "logicalType": "timestamp-micros",
            # Commented out since Avro's Python implementation (1.11.0) does not support local-timestamp-micros, even though it exists in the spec.
            # See bug report: https://issues.apache.org/jira/browse/AVRO-3476 and PR https://github.com/apache/avro/pull/1634
            # "logicalType": "timestamp-micros"
            # if timestamp_type.adjust_to_utc
            # else "local-timestamp-micros",
            "native_data_type": str(timestamp_type),
        }

    # visit_timestamptz() is required when using pyiceberg >= 0.5.0, which is essentially a duplicate
    # of visit_timestampz().  The function has been renamed from visit_timestampz().
    # Once Datahub can upgrade its pyiceberg dependency to >=0.5.0, the visit_timestampz() function can be safely removed.
    def visit_timestamptz(self, timestamptz_type: TimestamptzType) -> Dict[str, Any]:
        # Avro supports 2 types of timestamp:
        #  - Timestamp: independent of a particular timezone or calendar (TZ information is lost)
        #  - Local Timestamp: represents a timestamp in a local timezone, regardless of what specific time zone is considered local
        # utcAdjustment: bool = True
        return {
            "type": "long",
            "logicalType": "timestamp-micros",
            # Commented out since Avro's Python implementation (1.11.0) does not support local-timestamp-micros, even though it exists in the spec.
            # See bug report: https://issues.apache.org/jira/browse/AVRO-3476 and PR https://github.com/apache/avro/pull/1634
            # "logicalType": "timestamp-micros"
            # if timestamp_type.adjust_to_utc
            # else "local-timestamp-micros",
            "native_data_type": str(timestamptz_type),
        }

    def visit_timestampz(self, timestamptz_type: TimestamptzType) -> Dict[str, Any]:
        # Avro supports 2 types of timestamp:
        #  - Timestamp: independent of a particular timezone or calendar (TZ information is lost)
        #  - Local Timestamp: represents a timestamp in a local timezone, regardless of what specific time zone is considered local
        # utcAdjustment: bool = True
        return {
            "type": "long",
            "logicalType": "timestamp-micros",
            # Commented out since Avro's Python implementation (1.11.0) does not support local-timestamp-micros, even though it exists in the spec.
            # See bug report: https://issues.apache.org/jira/browse/AVRO-3476 and PR https://github.com/apache/avro/pull/1634
            # "logicalType": "timestamp-micros"
            # if timestamp_type.adjust_to_utc
            # else "local-timestamp-micros",
            "native_data_type": str(timestamptz_type),
        }

    def visit_string(self, string_type: StringType) -> Dict[str, Any]:
        return {
            "type": "string",
            "native_data_type": str(string_type),
        }

    def visit_uuid(self, uuid_type: UUIDType) -> Dict[str, Any]:
        return {
            "type": "string",
            "logicalType": "uuid",
            "native_data_type": str(uuid_type),
        }

    def visit_binary(self, binary_type: BinaryType) -> Dict[str, Any]:
        return {
            "type": "bytes",
            "native_data_type": str(binary_type),
        }



import logging
import textwrap
from dataclasses import dataclass
from typing import Iterable, List, Optional, Tuple

from pydantic import Field, SecretStr
from slack_sdk import WebClient

import datahub.emitter.mce_builder as builder
from datahub.configuration.common import ConfigModel
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.metadata.schema_classes import (
    CorpUserEditableInfoClass,
    DatasetPropertiesClass,
    DeprecationClass,
    SubTypesClass,
)
from datahub.utilities.ratelimiter import RateLimiter
from datahub.utilities.urns.urn import Urn

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class CorpUser:
    urn: Optional[str] = None
    email: Optional[str] = None
    slack_id: Optional[str] = None
    title: Optional[str] = None
    image_url: Optional[str] = None
    phone: Optional[str] = None
    real_name: Optional[str] = None
    slack_display_name: Optional[str] = None


class SlackSourceConfig(ConfigModel):
    bot_token: SecretStr = Field(
        description="Bot token for the Slack workspace. Needs `users:read`, `users:read.email` and `users.profile:read` scopes.",
    )
    enrich_user_metadata: bool = Field(
        type=bool,
        default=True,
        description="Whether to enrich user metadata.",
    )
    api_requests_per_min: int = Field(
        type=int,
        default=10,
        description="Number of API requests per minute. Low-level config. Do not tweak unless you are facing any issues.",
    )
    ingest_public_channels = Field(
        type=bool,
        default=False,
        description="Whether to ingest public channels. If set to true needs `channels:read` scope.",
    )
    channels_iteration_limit = Field(
        type=int,
        default=200,
        description="Limit the number of channels to be ingested in a iteration. Low-level config. Do not tweak unless you are facing any issues.",
    )
    channel_min_members = Field(
        type=int,
        default=2,
        description="Ingest channels with at least this many members.",
    )
    should_ingest_archived_channels = Field(
        type=bool,
        default=False,
        description="Whether to ingest archived channels.",
    )


@dataclass
class SlackSourceReport(SourceReport):
    channels_reported: int = 0
    archived_channels_reported: int = 0


PLATFORM_NAME = "slack"


@platform_name("Slack")
@config_class(SlackSourceConfig)
@support_status(SupportStatus.TESTING)
class SlackSource(Source):
    def __init__(self, ctx: PipelineContext, config: SlackSourceConfig):
        self.ctx = ctx
        self.config = config
        self.report = SlackSourceReport()
        self.workspace_base_url: Optional[str] = None
        self.rate_limiter = RateLimiter(
            max_calls=self.config.api_requests_per_min, period=60
        )
        self._use_users_info = False

    @classmethod
    def create(cls, config_dict, ctx):
        config = SlackSourceConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_slack_client(self) -> WebClient:
        return WebClient(token=self.config.bot_token.get_secret_value())

    def get_workunits_internal(
        self,
    ) -> Iterable[MetadataWorkUnit]:
        assert self.ctx.graph is not None
        auth_resp = self.get_slack_client().auth_test()
        assert isinstance(auth_resp.data, dict)
        self.workspace_base_url = str(auth_resp.data.get("url"))
        logger.info("Successfully connected to Slack")
        logger.info(auth_resp.data)
        if self.config.ingest_public_channels:
            yield from self.get_public_channels()
        if self.config.enrich_user_metadata:
            yield from self.get_user_info()

    def get_user_info(self) -> Iterable[MetadataWorkUnit]:
        assert self.ctx.graph is not None
        for user_obj in self.get_user_to_be_updated():
            self.populate_slack_id_from_email(user_obj)
            if user_obj.slack_id is None:
                continue
            self.populate_user_profile(user_obj)
            if user_obj.urn is None:
                continue
            logger.info(f"User: {user_obj}")
            corpuser_editable_info = (
                self.ctx.graph.get_aspect(
                    entity_urn=user_obj.urn, aspect_type=CorpUserEditableInfoClass
                )
                or CorpUserEditableInfoClass()
            )
            corpuser_editable_info.email = user_obj.email
            corpuser_editable_info.slack = user_obj.slack_id
            corpuser_editable_info.title = user_obj.title
            if user_obj.image_url:
                corpuser_editable_info.pictureLink = user_obj.image_url
            if user_obj.phone:
                corpuser_editable_info.phone = user_obj.phone
            if (
                not corpuser_editable_info.displayName
                or corpuser_editable_info.displayName == corpuser_editable_info.email
            ):
                # let's fill out a real name
                corpuser_editable_info.displayName = user_obj.real_name
            yield MetadataWorkUnit(
                id=f"{user_obj.urn}",
                mcp=MetadataChangeProposalWrapper(
                    entityUrn=user_obj.urn,
                    aspect=corpuser_editable_info,
                ),
            )

    def _get_channel_info(
        self, cursor: Optional[str]
    ) -> Tuple[List[MetadataWorkUnit], Optional[str]]:
        result_channels: List[MetadataWorkUnit] = []
        with self.rate_limiter:
            response = self.get_slack_client().conversations_list(
                types="public_channel",
                limit=self.config.channels_iteration_limit,
                cursor=cursor,
            )
        assert isinstance(response.data, dict)
        if not response.data["ok"]:
            self.report.report_failure(
                "public_channel", "Failed to fetch public channels"
            )
            return result_channels, None
        for channel in response.data["channels"]:
            num_members = channel["num_members"]
            if num_members < self.config.channel_min_members:
                continue
            channel_id = channel["id"]
            urn_channel = builder.make_dataset_urn(
                platform=PLATFORM_NAME, name=channel_id
            )
            name = channel["name"]
            is_archived = channel.get("is_archived", False)
            if is_archived:
                if not self.config.should_ingest_archived_channels:
                    continue
                self.report.archived_channels_reported += 1
                logger.info(f"Archived channel: {name}")
                result_channels.append(
                    MetadataWorkUnit(
                        id=f"{urn_channel}-deprecation",
                        mcp=MetadataChangeProposalWrapper(
                            entityUrn=urn_channel,
                            aspect=DeprecationClass(
                                deprecated=True,
                                note="This channel is archived",
                                actor="urn:li:corpuser:datahub",
                            ),
                        ),
                    )
                )

            topic = channel.get("topic", {}).get("value")
            purpose = channel.get("purpose", {}).get("value")
            if self.workspace_base_url:
                external_url = f"{self.workspace_base_url}/archives/{channel_id}"
            result_channels.append(
                MetadataWorkUnit(
                    id=f"{urn_channel}-datasetproperties",
                    mcp=MetadataChangeProposalWrapper(
                        entityUrn=urn_channel,
                        aspect=DatasetPropertiesClass(
                            name=name,
                            externalUrl=external_url,
                            description=f"Topic: {topic}\nPurpose: {purpose}",
                        ),
                    ),
                )
            )
            result_channels.append(
                MetadataWorkUnit(
                    id=f"{urn_channel}-subtype",
                    mcp=MetadataChangeProposalWrapper(
                        entityUrn=urn_channel,
                        aspect=SubTypesClass(
                            typeNames=["Slack Channel"],
                        ),
                    ),
                )
            )
        cursor = str(response.data["response_metadata"]["next_cursor"])
        return result_channels, cursor

    def get_public_channels(self) -> Iterable[MetadataWorkUnit]:
        cursor = None
        while True:
            with self.rate_limiter:
                channels, cursor = self._get_channel_info(cursor=cursor)
            for channel in channels:
                self.report.channels_reported += 1
                yield channel
            if not cursor:
                break

    def populate_user_profile(self, user_obj: CorpUser) -> None:
        if not user_obj.slack_id:
            return
        try:
            # https://api.slack.com/methods/users.profile.get
            with self.rate_limiter:
                if self._use_users_info:
                    user_profile_res = self.get_slack_client().users_info(
                        user=user_obj.slack_id
                    )
                    user_profile_res = user_profile_res.get("user", {})
                else:
                    user_profile_res = self.get_slack_client().users_profile_get(
                        user=user_obj.slack_id
                    )
            logger.debug(f"User profile: {user_profile_res}")
            user_profile = user_profile_res.get("profile", {})
            user_obj.title = user_profile.get("title")
            user_obj.image_url = user_profile.get("image_192")
            user_obj.phone = user_profile.get("phone")
            user_obj.real_name = user_profile.get("real_name")
            user_obj.slack_display_name = user_profile.get("display_name")

        except Exception as e:
            if "missing_scope" in str(e):
                if self._use_users_info:
                    raise e
                self._use_users_info = True
                self.populate_user_profile(user_obj)
            return

    def populate_slack_id_from_email(self, user_obj: CorpUser) -> None:
        if user_obj.email is None:
            return
        try:
            # https://api.slack.com/methods/users.lookupByEmail
            with self.rate_limiter:
                user_info_res = self.get_slack_client().users_lookupByEmail(
                    email=user_obj.email
                )
            user_info = user_info_res.get("user", {})
            user_obj.slack_id = user_info.get("id")
        except Exception as e:
            if "users_not_found" in str(e):
                return
            raise e

    def get_user_to_be_updated(self) -> Iterable[CorpUser]:
        graphql_query = textwrap.dedent(
            """
            query listUsers($input: ListUsersInput!) {
                listUsers(input: $input) {
                    total
                    users {
                        urn
                        editableProperties {
                            email
                            slack
                        }
                    }
                }
            }
        """
        )
        start = 0
        count = 10
        total = count

        assert self.ctx.graph is not None

        while start < total:
            variables = {"input": {"start": start, "count": count}}
            response = self.ctx.graph.execute_graphql(
                query=graphql_query, variables=variables
            )
            list_users = response.get("listUsers", {})
            total = list_users.get("total", 0)
            users = list_users.get("users", [])
            for user in users:
                user_obj = CorpUser()
                editable_properties = user.get("editableProperties", {})
                user_obj.urn = user.get("urn")
                if user_obj.urn is None:
                    continue
                if editable_properties is not None:
                    user_obj.email = editable_properties.get("email")
                if user_obj.email is None:
                    urn_id = Urn.from_string(user_obj.urn).get_entity_id_as_string()
                    if "@" in urn_id:
                        user_obj.email = urn_id
                if user_obj.email is not None:
                    yield user_obj
            start += count

    def get_report(self) -> SourceReport:
        return self.report



import functools
import importlib.resources as pkg_resource
import logging
import os
from typing import Dict, List

import lark
from lark import Lark, Tree

from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.source.powerbi.config import (
    PowerBiDashboardSourceConfig,
    PowerBiDashboardSourceReport,
)
from datahub.ingestion.source.powerbi.dataplatform_instance_resolver import (
    AbstractDataPlatformInstanceResolver,
)
from datahub.ingestion.source.powerbi.m_query import resolver, validator
from datahub.ingestion.source.powerbi.m_query.data_classes import (
    TRACE_POWERBI_MQUERY_PARSER,
)
from datahub.ingestion.source.powerbi.rest_api_wrapper.data_classes import Table
from datahub.utilities.threading_timeout import TimeoutException, threading_timeout

logger = logging.getLogger(__name__)

_M_QUERY_PARSE_TIMEOUT = int(os.getenv("DATAHUB_POWERBI_M_QUERY_PARSE_TIMEOUT", 60))


@functools.lru_cache(maxsize=1)
def get_lark_parser() -> Lark:
    # Read lexical grammar as text
    grammar: str = pkg_resource.read_text(
        "datahub.ingestion.source.powerbi", "powerbi-lexical-grammar.rule"
    )
    # Create lark parser for the grammar text
    return Lark(grammar, start="let_expression", regex=True)


def _parse_expression(expression: str) -> Tree:
    lark_parser: Lark = get_lark_parser()

    # Replace U+00a0 NO-BREAK SPACE with a normal space.
    # Sometimes PowerBI returns expressions with this character and it breaks the parser.
    expression = expression.replace("\u00a0", " ")

    logger.debug(f"Parsing expression = {expression}")
    with threading_timeout(_M_QUERY_PARSE_TIMEOUT):
        parse_tree: Tree = lark_parser.parse(expression)

    if TRACE_POWERBI_MQUERY_PARSER:
        logger.debug(parse_tree.pretty())

    return parse_tree


def get_upstream_tables(
    table: Table,
    reporter: PowerBiDashboardSourceReport,
    platform_instance_resolver: AbstractDataPlatformInstanceResolver,
    ctx: PipelineContext,
    config: PowerBiDashboardSourceConfig,
    parameters: Dict[str, str] = {},
) -> List[resolver.Lineage]:

    if table.expression is None:
        logger.debug(f"There is no M-Query expression in table {table.full_name}")
        return []

    parameters = parameters or {}

    logger.debug(
        f"Processing {table.full_name} m-query expression for lineage extraction. Expression = {table.expression}"
    )

    try:
        with reporter.m_query_parse_timer:
            reporter.m_query_parse_attempts += 1
            parse_tree: Tree = _parse_expression(table.expression)

        valid, message = validator.validate_parse_tree(
            parse_tree, native_query_enabled=config.native_query_parsing
        )
        if valid is False:
            assert message is not None
            logger.debug(f"Validation failed: {message}")
            reporter.info(
                title="Unsupported M-Query",
                message="DataAccess function is not present in M-Query expression",
                context=f"table-full-name={table.full_name}, expression={table.expression}, message={message}",
            )
            reporter.m_query_parse_validation_errors += 1
            return []
    except KeyboardInterrupt:
        raise
    except TimeoutException:
        reporter.m_query_parse_timeouts += 1
        reporter.warning(
            title="M-Query Parsing Timeout",
            message=f"M-Query parsing timed out after {_M_QUERY_PARSE_TIMEOUT} seconds. Lineage for this table will not be extracted.",
            context=f"table-full-name={table.full_name}, expression={table.expression}",
        )
        return []
    except (
        BaseException
    ) as e:  # TODO: Debug why BaseException is needed here and below.
        if isinstance(e, lark.exceptions.UnexpectedCharacters):
            error_type = "Unexpected Character Error"
            reporter.m_query_parse_unexpected_character_errors += 1
        else:
            error_type = "Unknown Parsing Error"
            reporter.m_query_parse_unknown_errors += 1

        reporter.warning(
            title="Unable to extract lineage from M-Query expression",
            message=f"Got an '{error_type}' while parsing the expression. Lineage will be missing for this table.",
            context=f"table-full-name={table.full_name}, expression={table.expression}",
            exc=e,
        )
        return []
    reporter.m_query_parse_successes += 1

    try:
        lineage: List[resolver.Lineage] = resolver.MQueryResolver(
            table=table,
            parse_tree=parse_tree,
            reporter=reporter,
            parameters=parameters,
        ).resolve_to_data_platform_table_list(
            ctx=ctx,
            config=config,
            platform_instance_resolver=platform_instance_resolver,
        )

        return lineage

    except BaseException as e:
        reporter.m_query_resolver_errors += 1
        reporter.warning(
            title="Unknown M-Query Pattern",
            message="Encountered a unknown M-Query Expression",
            context=f"table-full-name={table.full_name}, expression={table.expression}, message={e}",
            exc=e,
        )
        return []



import atexit
import functools
import logging
import os
from typing import Iterable, List, Optional

from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.incremental_lineage_helper import auto_incremental_lineage
from datahub.ingestion.api.source import (
    MetadataWorkUnitProcessor,
    SourceCapability,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.bigquery_v2.bigquery_audit import BigqueryTableIdentifier
from datahub.ingestion.source.bigquery_v2.bigquery_config import BigQueryV2Config
from datahub.ingestion.source.bigquery_v2.bigquery_report import BigQueryV2Report
from datahub.ingestion.source.bigquery_v2.bigquery_schema import (
    BigQuerySchemaApi,
    get_projects,
)
from datahub.ingestion.source.bigquery_v2.bigquery_schema_gen import (
    BigQuerySchemaGenerator,
)
from datahub.ingestion.source.bigquery_v2.bigquery_test_connection import (
    BigQueryTestConnection,
)
from datahub.ingestion.source.bigquery_v2.common import (
    BigQueryFilter,
    BigQueryIdentifierBuilder,
)
from datahub.ingestion.source.bigquery_v2.lineage import BigqueryLineageExtractor
from datahub.ingestion.source.bigquery_v2.profiler import BigqueryProfiler
from datahub.ingestion.source.bigquery_v2.queries_extractor import (
    BigQueryQueriesExtractor,
    BigQueryQueriesExtractorConfig,
)
from datahub.ingestion.source.bigquery_v2.usage import BigQueryUsageExtractor
from datahub.ingestion.source.state.profiling_state_handler import ProfilingHandler
from datahub.ingestion.source.state.redundant_run_skip_handler import (
    RedundantLineageRunSkipHandler,
    RedundantUsageRunSkipHandler,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.ingestion.source_report.ingestion_stage import QUERIES_EXTRACTION
from datahub.sql_parsing.schema_resolver import SchemaResolver
from datahub.utilities.registries.domain_registry import DomainRegistry

logger: logging.Logger = logging.getLogger(__name__)


# We can't use close as it is not called if the ingestion is not successful
def cleanup(config: BigQueryV2Config) -> None:
    if config._credentials_path is not None:
        os.unlink(config._credentials_path)


@platform_name("BigQuery", doc_order=1)
@config_class(BigQueryV2Config)
@support_status(SupportStatus.CERTIFIED)
@capability(  # DataPlatformAspect is set to project id, but not added to urns as project id is in the container path
    SourceCapability.PLATFORM_INSTANCE,
    "Platform instance is pre-set to the BigQuery project id",
    supported=False,
)
@capability(SourceCapability.DOMAINS, "Supported via the `domain` config field")
@capability(SourceCapability.CONTAINERS, "Enabled by default")
@capability(SourceCapability.SCHEMA_METADATA, "Enabled by default")
@capability(
    SourceCapability.DATA_PROFILING,
    "Optionally enabled via configuration",
)
@capability(SourceCapability.DESCRIPTIONS, "Enabled by default")
@capability(SourceCapability.LINEAGE_COARSE, "Optionally enabled via configuration")
@capability(SourceCapability.LINEAGE_FINE, "Optionally enabled via configuration")
@capability(
    SourceCapability.USAGE_STATS,
    "Enabled by default, can be disabled via configuration `include_usage_statistics`",
)
@capability(
    SourceCapability.CLASSIFICATION,
    "Optionally enabled via `classification.enabled`",
    supported=True,
)
class BigqueryV2Source(StatefulIngestionSourceBase, TestableSource):
    def __init__(self, ctx: PipelineContext, config: BigQueryV2Config):
        super().__init__(config, ctx)
        self.config: BigQueryV2Config = config
        self.report: BigQueryV2Report = BigQueryV2Report()
        self.platform: str = "bigquery"

        self.domain_registry: Optional[DomainRegistry] = None
        if self.config.domain:
            self.domain_registry = DomainRegistry(
                cached_domains=[k for k in self.config.domain], graph=self.ctx.graph
            )

        BigqueryTableIdentifier._BIGQUERY_DEFAULT_SHARDED_TABLE_REGEX = (
            self.config.sharded_table_pattern
        )

        self.bigquery_data_dictionary = BigQuerySchemaApi(
            report=self.report.schema_api_perf,
            projects_client=config.get_projects_client(),
            client=config.get_bigquery_client(),
        )
        if self.config.extract_policy_tags_from_catalog:
            self.bigquery_data_dictionary.datacatalog_client = (
                self.config.get_policy_tag_manager_client()
            )

        with self.report.init_schema_resolver_timer:
            self.sql_parser_schema_resolver = self._init_schema_resolver()

        self.filters = BigQueryFilter(self.config, self.report)
        self.identifiers = BigQueryIdentifierBuilder(self.config, self.report)

        redundant_lineage_run_skip_handler: Optional[
            RedundantLineageRunSkipHandler
        ] = None
        if self.config.enable_stateful_lineage_ingestion:
            redundant_lineage_run_skip_handler = RedundantLineageRunSkipHandler(
                source=self,
                config=self.config,
                pipeline_name=self.ctx.pipeline_name,
                run_id=self.ctx.run_id,
            )

        # For database, schema, tables, views, snapshots etc
        self.lineage_extractor = BigqueryLineageExtractor(
            config,
            self.report,
            schema_resolver=self.sql_parser_schema_resolver,
            identifiers=self.identifiers,
            redundant_run_skip_handler=redundant_lineage_run_skip_handler,
        )

        redundant_usage_run_skip_handler: Optional[RedundantUsageRunSkipHandler] = None
        if self.config.enable_stateful_usage_ingestion:
            redundant_usage_run_skip_handler = RedundantUsageRunSkipHandler(
                source=self,
                config=self.config,
                pipeline_name=self.ctx.pipeline_name,
                run_id=self.ctx.run_id,
            )

        self.usage_extractor = BigQueryUsageExtractor(
            config,
            self.report,
            schema_resolver=self.sql_parser_schema_resolver,
            identifiers=self.identifiers,
            redundant_run_skip_handler=redundant_usage_run_skip_handler,
        )

        self.profiling_state_handler: Optional[ProfilingHandler] = None
        if self.config.enable_stateful_profiling:
            self.profiling_state_handler = ProfilingHandler(
                source=self,
                config=self.config,
                pipeline_name=self.ctx.pipeline_name,
                run_id=self.ctx.run_id,
            )
        self.profiler = BigqueryProfiler(
            config, self.report, self.profiling_state_handler
        )

        self.bq_schema_extractor = BigQuerySchemaGenerator(
            self.config,
            self.report,
            self.bigquery_data_dictionary,
            self.domain_registry,
            self.sql_parser_schema_resolver,
            self.profiler,
            self.identifiers,
            self.ctx.graph,
        )

        self.add_config_to_report()
        atexit.register(cleanup, config)

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "BigqueryV2Source":
        config = BigQueryV2Config.parse_obj(config_dict)
        return cls(ctx, config)

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        return BigQueryTestConnection.test_connection(config_dict)

    def _init_schema_resolver(self) -> SchemaResolver:
        schema_resolution_required = (
            self.config.use_queries_v2
            or self.config.lineage_parse_view_ddl
            or self.config.lineage_use_sql_parser
        )
        schema_ingestion_enabled = (
            self.config.include_schema_metadata
            and self.config.include_tables
            and self.config.include_views
            and self.config.include_table_snapshots
        )

        if schema_resolution_required and not schema_ingestion_enabled:
            if self.ctx.graph:
                return self.ctx.graph.initialize_schema_resolver_from_datahub(
                    platform=self.platform,
                    platform_instance=self.config.platform_instance,
                    env=self.config.env,
                    batch_size=self.config.schema_resolution_batch_size,
                )
            else:
                logger.warning(
                    "Failed to load schema info from DataHub as DataHubGraph is missing. "
                    "Use `datahub-rest` sink OR provide `datahub-api` config in recipe. ",
                )
        return SchemaResolver(platform=self.platform, env=self.config.env)

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            functools.partial(
                auto_incremental_lineage, self.config.incremental_lineage
            ),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        projects = get_projects(
            self.bq_schema_extractor.schema_api,
            self.report,
            self.filters,
        )
        if not projects:
            return

        for project in projects:
            yield from self.bq_schema_extractor.get_project_workunits(project)

        if self.config.use_queries_v2:
            # Always ingest View and Snapshot lineage with schema ingestion
            self.report.set_ingestion_stage("*", "View and Snapshot Lineage")

            yield from self.lineage_extractor.get_lineage_workunits_for_views_and_snapshots(
                [p.id for p in projects],
                self.bq_schema_extractor.view_refs_by_project,
                self.bq_schema_extractor.view_definitions,
                self.bq_schema_extractor.snapshot_refs_by_project,
                self.bq_schema_extractor.snapshots_by_ref,
            )

            # if both usage and lineage are disabled then skip queries extractor piece
            if (
                not self.config.include_usage_statistics
                and not self.config.include_table_lineage
            ):
                return

            self.report.set_ingestion_stage("*", QUERIES_EXTRACTION)

            with BigQueryQueriesExtractor(
                connection=self.config.get_bigquery_client(),
                schema_api=self.bq_schema_extractor.schema_api,
                config=BigQueryQueriesExtractorConfig(
                    window=self.config,
                    user_email_pattern=self.config.usage.user_email_pattern,
                    include_lineage=self.config.include_table_lineage,
                    include_usage_statistics=self.config.include_usage_statistics,
                    include_operations=self.config.usage.include_operational_stats,
                    top_n_queries=self.config.usage.top_n_queries,
                    region_qualifiers=self.config.region_qualifiers,
                ),
                structured_report=self.report,
                filters=self.filters,
                identifiers=self.identifiers,
                schema_resolver=self.sql_parser_schema_resolver,
                discovered_tables=self.bq_schema_extractor.table_refs,
            ) as queries_extractor:
                self.report.queries_extractor = queries_extractor.report
                yield from queries_extractor.get_workunits_internal()

        else:
            if self.config.include_usage_statistics:
                yield from self.usage_extractor.get_usage_workunits(
                    [p.id for p in projects], self.bq_schema_extractor.table_refs
                )

            if self.config.include_table_lineage:
                yield from self.lineage_extractor.get_lineage_workunits(
                    [p.id for p in projects],
                    self.bq_schema_extractor.view_refs_by_project,
                    self.bq_schema_extractor.view_definitions,
                    self.bq_schema_extractor.snapshot_refs_by_project,
                    self.bq_schema_extractor.snapshots_by_ref,
                    self.bq_schema_extractor.table_refs,
                )

        # Lineage BQ to GCS
        if (
            self.config.include_table_lineage
            and self.bq_schema_extractor.external_tables
        ):
            for dataset_urn, table in self.bq_schema_extractor.external_tables.items():
                yield from self.lineage_extractor.gen_lineage_workunits_for_external_table(
                    dataset_urn, table.ddl, graph=self.ctx.graph
                )

    def get_report(self) -> BigQueryV2Report:
        return self.report

    def add_config_to_report(self):
        self.report.include_table_lineage = self.config.include_table_lineage
        self.report.use_date_sharded_audit_log_tables = (
            self.config.use_date_sharded_audit_log_tables
        )
        self.report.log_page_size = self.config.log_page_size
        self.report.use_exported_bigquery_audit_metadata = (
            self.config.use_exported_bigquery_audit_metadata
        )
        self.report.stateful_lineage_ingestion_enabled = (
            self.config.enable_stateful_lineage_ingestion
        )
        self.report.stateful_usage_ingestion_enabled = (
            self.config.enable_stateful_usage_ingestion
        )
        self.report.window_start_time, self.report.window_end_time = (
            self.config.start_time,
            self.config.end_time,
        )
